from typing import Union
import numpy as np
import pandas as pd
import multiprocessing as mp
from .kernel import GWRKernel, GTWRKernel
from .function import _compute_betas_gwr, surface_to_plane
from .obj import CalAicObj, CalMultiObj, BaseModel, GWRResults, GTWRResults, MGWRResults, MGTWRResults
from joblib import Parallel, delayed


class GWR(BaseModel):
    """
    Geographically Weighted Regression
    """
    def __init__(
            self,
            coords: Union[np.ndarray, pd.DataFrame],
            X: Union[np.ndarray, pd.DataFrame],
            y: Union[np.ndarray, pd.DataFrame, pd.Series],
            bw: float,
            kernel: str = 'bisquare',
            fixed: bool = True,
            constant: bool = True,
            thread: int = 1,
            convert: bool = False,
    ):
        """
        Parameters
        ----------
        coords        : array-like
                        n*2, spatial coordinates of the observations, if it's latitude and longitude,
                        the first column should be longitude

        X             : array-like
                        n*k, independent variable, excluding the constant

        y             : array-like
                        n*1, dependent variable

        bw            : scalar
                        bandwidth value consisting of either a distance or N
                        nearest neighbors; user specified or obtained using
                        sel

        kernel        : string
                        type of kernel function used to weight observations;
                        available options:
                        'gaussian'
                        'bisquare'
                        'exponential'

        fixed         : bool
                        True for distance based kernel function and  False for
                        adaptive (nearest neighbor) kernel function (default)

        constant      : bool
                        True to include intercept (default) in model and False to exclude
                        intercept.

        thread        : int
                        The number of processes in parallel computation. If you have a large amount of data,
                        you can use it

        convert       : bool
                        Whether to convert latitude and longitude to plane coordinates.
        Examples
        --------
        import numpy as np
        from mgtwr.model import GWR
        np.random.seed(10)
        u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
        v = np.array([((i-1) % 144) // 12 for i in range(1, 1729)]).reshape(-1, 1)
        t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
        x1 = np.random.uniform(0, 1, (1728, 1))
        x2 = np.random.uniform(0, 1, (1728, 1))
        epsilon = np.random.randn(1728, 1)
        beta0 = 5
        beta1 = 3 + (u + v + t)/6
        beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
        y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
        coords = np.hstack([u, v])
        X = np.hstack([x1, x2])
        gwr = GWR(coords, X, y, 0.8, kernel='gaussian', fixed=True).fit()
        print(gwr.R2)
        0.7128737240047688
        """
        super(GWR, self).__init__(X, y, kernel, fixed, constant)
        if thread < 1 or not isinstance(thread, int):
            raise ValueError('thread should be an integer greater than or equal to 1')
        if isinstance(coords, pd.DataFrame):
            coords = coords.values
        self.coords = coords
        if convert:
            longitude = coords[:, 0]
            latitude = coords[:, 1]
            longitude, latitude = surface_to_plane(longitude, latitude)
            self.coords = np.hstack([longitude, latitude])
        self.bw = bw
        self.thread = thread

    def _build_wi(self, i, bw):
        """
        calculate Weight matrix
        """
        try:
            gwr_kernel = GWRKernel(self.coords, bw, fixed=self.fixed, function=self.kernel)
            distance = gwr_kernel.cal_distance(i)
            wi = gwr_kernel.cal_kernel(distance)
        except BaseException:
            raise  # TypeError('Unsupported kernel function  ', kernel)

        return wi

    def cal_aic(self):
        """
        use for calculating AICc, BIC, CV and so on.
        """
        if self.thread > 1:
            result = list(zip(*Parallel(n_jobs=self.thread)(delayed(self._search_local_fit)(i) for i in range(self.n))))
        else:
            result = list(zip(*map(self._search_local_fit, range(self.n))))
        err2 = np.array(result[0]).reshape(-1, 1)
        hat = np.array(result[1]).reshape(-1, 1)
        aa = np.sum(err2 / ((1 - hat) ** 2))
        RSS = np.sum(err2)
        tr_S = np.sum(hat)
        llf = -np.log(RSS) * self.n / 2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2

        return CalAicObj(tr_S, float(llf), float(aa), self.n)

    def _search_local_fit(self, i):
        wi = self._build_wi(i, self.bw).reshape(-1, 1)
        betas, inv_xtx_xt = _compute_betas_gwr(self.y, self.X, wi)
        predict = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - predict
        influx = np.dot(self.X[i], inv_xtx_xt[:, i])
        return reside * reside, influx

    def _local_fit(self, i):
        wi = self._build_wi(i, self.bw).reshape(-1, 1)
        betas, inv_xtx_xt = _compute_betas_gwr(self.y, self.X, wi)
        predict = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - predict
        influx = np.dot(self.X[i], inv_xtx_xt[:, i])
        Si = np.dot(self.X[i], inv_xtx_xt).reshape(-1)
        CCT = np.diag(np.dot(inv_xtx_xt, inv_xtx_xt.T)).reshape(-1)
        Si2 = np.sum(Si ** 2)
        return influx, reside, predict, betas.reshape(-1), CCT, Si2

    def _multi_fit(self, i):
        wi = self._build_wi(i, self.bw).reshape(-1, 1)
        betas, inv_xtx_xt = _compute_betas_gwr(self.y, self.X, wi)
        pre = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - pre
        return betas.reshape(-1), pre, reside

    def cal_multi(self):
        """
        calculate betas, predict value and reside, use for searching best bandwidth in MGWR model by backfitting.
        """
        if self.thread > 1:
            result = list(zip(*Parallel(n_jobs=self.thread)(delayed(self._multi_fit)(i) for i in range(self.n))))
        else:
            result = list(zip(*map(self._multi_fit, range(self.n))))
        betas = np.array(result[0])
        pre = np.array(result[1]).reshape(-1, 1)
        reside = np.array(result[2]).reshape(-1, 1)
        return CalMultiObj(betas, pre, reside)

    def fit(self):
        """
        To fit GWR model
        """
        if self.thread > 1:
            result = list(zip(*Parallel(n_jobs=self.thread)(delayed(self._local_fit)(i) for i in range(self.n))))
        else:
            result = list(zip(*map(self._local_fit, range(self.n))))
        influ = np.array(result[0]).reshape(-1, 1)
        reside = np.array(result[1]).reshape(-1, 1)
        predict_value = np.array(result[2]).reshape(-1, 1)
        betas = np.array(result[3])
        CCT = np.array(result[4])
        tr_STS = np.array(result[5])
        return GWRResults(self.coords, self.X, self.y, self.bw, self.kernel, self.fixed,
                          influ, reside, predict_value, betas, CCT, tr_STS)


class MGWR(GWR):
    """
    Multiscale Geographically Weighted Regression
    """
    def __init__(
            self,
            coords: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            selector,
            kernel: str = 'bisquare',
            fixed: bool = False,
            constant: bool = True,
            thread: int = 1,
            convert: bool = False
    ):
        """
        Parameters
        ----------
        coords        : array-like
                        n*2, spatial coordinates of the observations, if it's latitude and longitude,
                        the first column should be longitude

        X             : array-like
                        n*k, independent variable, excluding the constant

        y             : array-like
                        n*1, dependent variable

        selector      :SearchMGWRParameter object
                       valid SearchMGWRParameter that has successfully called
                       the "search" method. This parameter passes on
                       information from GAM model estimation including optimal
                       bandwidths.

        kernel        : string
                        type of kernel function used to weight observations;
                        available options:
                        'gaussian'
                        'bisquare'
                        'exponential'

        fixed         : bool
                        True for distance based kernel function and  False for
                        adaptive (nearest neighbor) kernel function (default)

        constant      : bool
                        True to include intercept (default) in model and False to exclude
                        intercept.

        thread        : int
                        The number of processes in parallel computation. If you have a large amount of data,
                        you can use it

        convert       : bool
                        Whether to convert latitude and longitude to plane coordinates.
        Examples
        --------
        import numpy as np
        from mgtwr.sel import SearchMGWRParameter
        from mgtwr.model import MGWR
        np.random.seed(10)
        u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
        v = np.array([((i-1) % 144) // 12 for i in range(1, 1729)]).reshape(-1, 1)
        t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
        x1 = np.random.uniform(0, 1, (1728, 1))
        x2 = np.random.uniform(0, 1, (1728, 1))
        epsilon = np.random.randn(1728, 1)
        beta0 = 5
        beta1 = 3 + (u + v + t)/6
        beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
        y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
        coords = np.hstack([u, v])
        X = np.hstack([x1, x2])
        sel_multi = SearchMGWRParameter(coords, X, y, kernel='gaussian', fixed=True)
        bws = sel_multi.search(multi_bw_max=[40], verbose=True)
        mgwr = MGWR(coords, X, y, sel_multi, kernel='gaussian', fixed=True).fit()
        print(mgwr.R2)
        0.7045642214972343
        """
        self.selector = selector
        self.bws = self.selector.bws[0]  # final set of bandwidth
        self.bws_history = selector.bws[1]  # bws history in back_fitting
        self.betas = selector.bws[3]
        bw_init = self.selector.bws[5]  # initialization bandwidth
        super().__init__(
            coords, X, y, bw_init, kernel=kernel, fixed=fixed, constant=constant, thread=thread, convert=convert)
        self.n_chunks = None
        self.ENP_j = None

    def _chunk_compute(self, chunk_id=0):
        n = self.n
        k = self.k
        n_chunks = self.n_chunks
        chunk_size = int(np.ceil(float(n / n_chunks)))
        ENP_j = np.zeros(self.k)
        CCT = np.zeros((self.n, self.k))

        chunk_index = np.arange(n)[chunk_id * chunk_size:(chunk_id + 1) * chunk_size]
        init_pR = np.zeros((n, len(chunk_index)))
        init_pR[chunk_index, :] = np.eye(len(chunk_index))
        pR = np.zeros((n, len(chunk_index),
                       k))  # partial R: n by chunk_size by k

        for i in range(n):
            wi = self._build_wi(i, self.bw).reshape(-1, 1)
            xT = (self.X * wi).T
            P = np.linalg.solve(xT.dot(self.X), xT).dot(init_pR).T
            pR[i, :, :] = P * self.X[i]

        err = init_pR - np.sum(pR, axis=2)  # n by chunk_size

        for iter_i in range(self.bws_history.shape[0]):
            for j in range(k):
                pRj_old = pR[:, :, j] + err
                Xj = self.X[:, j]
                n_chunks_Aj = n_chunks
                chunk_size_Aj = int(np.ceil(float(n / n_chunks_Aj)))
                for chunk_Aj in range(n_chunks_Aj):
                    chunk_index_Aj = np.arange(n)[chunk_Aj * chunk_size_Aj:(
                                                                                   chunk_Aj + 1) * chunk_size_Aj]
                    pAj = np.empty((len(chunk_index_Aj), n))
                    for i in range(len(chunk_index_Aj)):
                        index = chunk_index_Aj[i]
                        wi = self._build_wi(index, self.bws_history[iter_i, j])
                        xw = Xj * wi
                        pAj[i, :] = Xj[index] / np.sum(xw * Xj) * xw
                    pR[chunk_index_Aj, :, j] = pAj.dot(pRj_old)
                err = pRj_old - pR[:, :, j]

        for j in range(k):
            CCT[:, j] += ((pR[:, :, j] / self.X[:, j].reshape(-1, 1)) ** 2).sum(
                axis=1)
        for i in range(len(chunk_index)):
            ENP_j += pR[chunk_index[i], i, :]

        return ENP_j, CCT,

    def fit(self, n_chunks=1):
        """
        Compute MGWR inference by chunk to reduce memory footprint.
        """
        self.n_chunks = n_chunks
        pre = np.sum(self.X * self.betas, axis=1).reshape(-1, 1)
        result = map(self._chunk_compute, (range(n_chunks)))
        result_list = list(zip(*result))
        ENP_j = np.sum(np.array(result_list[0]), axis=0)
        CCT = np.sum(np.array(result_list[1]), axis=0)
        return MGWRResults(
            self.coords, self.X, self.y, self.bws, self.kernel, self.fixed,
            self.bws_history, self.betas, pre, ENP_j, CCT)


class GTWR(BaseModel):
    """
    Geographically and Temporally Weighted Regression

    Parameters
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates of
                    observations

    t             : array-like
                    n*1, time location

    X             : array-like
                        n*k, independent variable, excluding the constant

    y             : array-like
                    n*1, dependent variable

    bw            : scalar
                    bandwidth value consisting of either a distance or N
                    nearest neighbors; user specified or obtained using
                    sel

    tau           : scalar
                    spatio-temporal scale

    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'

    fixed         : bool
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)

    constant      : bool
                    True to include intercept (default) in model and False to exclude
                    intercept.

    Examples
    --------
    import numpy as np
    from mgtwr.model import GTWR
    np.random.seed(10)
    u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
    v = np.array([((i-1) % 144) // 12 for i in range(1, 1729)]).reshape(-1, 1)
    t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
    x1 = np.random.uniform(0, 1, (1728, 1))
    x2 = np.random.uniform(0, 1, (1728, 1))
    epsilon = np.random.randn(1728, 1)
    beta0 = 5
    beta1 = 3 + (u + v + t)/6
    beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
    y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
    coords = np.hstack([u, v])
    X = np.hstack([x1, x2])
    gtwr = GTWR(coords, t, X, y, 0.8, 1.9, kernel='gaussian', fixed=True).fit()
    print(gtwr.R2)
    0.9899869616636376
    """

    def __init__(
            self,
            coords: Union[np.ndarray, pd.DataFrame],
            t: Union[np.ndarray, pd.DataFrame],
            X: Union[np.ndarray, pd.DataFrame],
            y: Union[np.ndarray, pd.DataFrame],
            bw: float,
            tau: float,
            kernel: str = 'gaussian',
            fixed: bool = False,
            constant: bool = True,
            thread: int = 1,
            convert: bool = False
    ):
        super(GTWR, self).__init__(X, y, kernel, fixed, constant)
        if thread < 1 or not isinstance(thread, int):
            raise ValueError('thread should be an integer greater than or equal to 1')
        if isinstance(coords, pd.DataFrame):
            coords = coords.values
        self.coords = coords
        if convert:
            longitude = coords[:, 0]
            latitude = coords[:, 1]
            longitude, latitude = surface_to_plane(longitude, latitude)
            self.coords = np.hstack([longitude, latitude])
        self.t = t
        self.bw = bw
        self.tau = tau
        self.bw_s = self.bw
        self.bw_t = np.sqrt(self.bw ** 2 / self.tau)
        self.thread = thread

    def _build_wi(self, i, bw, tau):
        """
        calculate Weight matrix
        """
        try:
            gtwr_kernel = GTWRKernel(self.coords, self.t, bw, tau, fixed=self.fixed, function=self.kernel)
            distance = gtwr_kernel.cal_distance(i)
            wi = gtwr_kernel.cal_kernel(distance)
        except BaseException:
            raise  # TypeError('Unsupported kernel function  ', kernel)

        return wi

    def cal_aic(self):

        """
        use for calculating AICc, BIC, CV and so on.
        """
        if self.thread > 1:
            result = list(zip(*Parallel(n_jobs=self.thread)(delayed(self._search_local_fit)(i) for i in range(self.n))))
        else:
            result = list(zip(*map(self._search_local_fit, range(self.n))))
        err2 = np.array(result[0]).reshape(-1, 1)
        hat = np.array(result[1]).reshape(-1, 1)
        aa = np.sum(err2 / ((1 - hat) ** 2))
        RSS = np.sum(err2)
        tr_S = np.sum(hat)
        llf = -np.log(RSS) * self.n / 2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2

        return CalAicObj(tr_S, float(llf), float(aa), self.n)

    def _search_local_fit(self, i):
        wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
        betas, xtx_inv_xt = _compute_betas_gwr(self.y, self.X, wi)
        predict = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - predict
        influ = np.dot(self.X[i], xtx_inv_xt[:, i])
        return reside * reside, influ

    def _local_fit(self, i):
        wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
        betas, xtx_inv_xt = _compute_betas_gwr(self.y, self.X, wi)
        predict = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - predict
        influ = np.dot(self.X[i], xtx_inv_xt[:, i])
        Si = np.dot(self.X[i], xtx_inv_xt).reshape(-1)
        CCT = np.diag(np.dot(xtx_inv_xt, xtx_inv_xt.T)).reshape(-1)
        Si2 = np.sum(Si ** 2)
        return influ, reside, predict, betas.reshape(-1), CCT, Si2

    def _multi_fit(self, i):
        wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
        betas, inv_xtx_xt = _compute_betas_gwr(self.y, self.X, wi)
        pre = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - pre
        return betas.reshape(-1), pre, reside

    def cal_multi(self):
        """
        calculate betas, predict value and reside, use for searching best bandwidth in MGWR model by backfitting.
        """
        if self.thread > 1:
            result = list(zip(*Parallel(n_jobs=self.thread)(delayed(self._multi_fit)(i) for i in range(self.n))))
        else:
            result = list(zip(*map(self._multi_fit, range(self.n))))
        betas = np.array(result[0])
        pre = np.array(result[1]).reshape(-1, 1)
        reside = np.array(result[2]).reshape(-1, 1)
        return CalMultiObj(betas, pre, reside)

    def fit(self):
        """
        fit GTWR models

        """
        if self.thread > 1:
            result = list(zip(*Parallel(n_jobs=self.thread)(delayed(self._local_fit)(i) for i in range(self.n))))
        else:
                result = list(zip(*map(self._local_fit, range(self.n))))
        influ = np.array(result[0]).reshape(-1, 1)
        reside = np.array(result[1]).reshape(-1, 1)
        predict_value = np.array(result[2]).reshape(-1, 1)
        betas = np.array(result[3])
        CCT = np.array(result[4])
        tr_STS = np.array(result[5])
        return GTWRResults(
            self.coords, self.t, self.X, self.y, self.bw, self.tau, self.kernel, self.fixed,
            influ, reside, predict_value, betas, CCT, tr_STS
        )


class MGTWR(GTWR):
    """
    Multiscale GTWR estimation and inference.

    Parameters
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates of
                    observatons

    t             : array
                    n*1, time location

    X             : array-like
                        n*k, independent variable, excluding the constant

    y             : array-like
                    n*1, dependent variable

    selector      : SearchMGTWRParameter object
                    valid SearchMGTWRParameter object that has successfully called
                    the "search" method. This parameter passes on
                    information from GAM model estimation including optimal
                    bandwidths.

    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'

    fixed         : bool
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)

    constant      : bool
                    True to include intercept (default) in model and False to exclude
                    intercept.
    Examples
    --------
    import numpy as np
    from mgtwr.sel import SearchMGTWRParameter
    from mgtwr.model import MGTWR
    np.random.seed(10)
    u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
    v = np.array([((i-1) % 144)//12 for i in range(1, 1729)]).reshape(-1, 1)
    t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
    x1 = np.random.uniform(0, 1, (1728, 1))
    x2 = np.random.uniform(0, 1, (1728, 1))
    epsilon = np.random.randn(1728, 1)
    beta0 = 5
    beta1 = 3 + (u + v + t)/6
    beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
    y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
    coords = np.hstack([u, v])
    X = np.hstack([x1, x2])
    sel_multi = SearchMGTWRParameter(coords, t, X, y, kernel='gaussian', fixed=True)
    bws = sel_multi.search(multi_bw_min=[0.1], verbose=True, tol_multi=1.0e-4)
    mgtwr = MGTWR(coords, t, X, y, sel_multi, kernel='gaussian', fixed=True).fit()
    print(mgtwr.R2)
    0.9972924820674222
    """

    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            selector,
            kernel: str = 'bisquare',
            fixed: bool = False,
            constant: bool = True,
            thread: int = 1,
            convert: bool = False
    ):
        self.selector = selector
        self.bws = self.selector.bws[0]  # final set of bandwidth
        self.taus = self.selector.bws[1]
        self.bw_ts = np.sqrt(self.bws ** 2 / self.taus)
        self.bws_history = selector.bws[2]  # bws history in back_fitting
        self.taus_history = selector.bws[3]
        self.betas = selector.bws[5]
        bw_init = self.selector.bws[7]  # initialization bandwidth
        tau_init = self.selector.bws[8]
        super().__init__(coords, t, X, y, bw_init, tau_init,
                         kernel=kernel, fixed=fixed, constant=constant, thread=thread, convert=convert)
        self.n_chunks = None
        self.ENP_j = None

    def _chunk_compute(self, chunk_id=0):
        n = self.n
        k = self.k
        n_chunks = self.n_chunks
        chunk_size = int(np.ceil(float(n / n_chunks)))
        ENP_j = np.zeros(self.k)
        CCT = np.zeros((self.n, self.k))

        chunk_index = np.arange(n)[chunk_id * chunk_size:(chunk_id + 1) * chunk_size]
        init_pR = np.zeros((n, len(chunk_index)))
        init_pR[chunk_index, :] = np.eye(len(chunk_index))
        pR = np.zeros((n, len(chunk_index),
                       k))  # partial R: n by chunk_size by k

        for i in range(n):
            wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
            xT = (self.X * wi).T
            P = np.linalg.solve(xT.dot(self.X), xT).dot(init_pR).T
            pR[i, :, :] = P * self.X[i]

        err = init_pR - np.sum(pR, axis=2)  # n by chunk_size

        for iter_i in range(self.bws_history.shape[0]):
            for j in range(k):
                pRj_old = pR[:, :, j] + err
                Xj = self.X[:, j]
                n_chunks_Aj = n_chunks
                chunk_size_Aj = int(np.ceil(float(n / n_chunks_Aj)))
                for chunk_Aj in range(n_chunks_Aj):
                    chunk_index_Aj = np.arange(n)[chunk_Aj * chunk_size_Aj:(
                                                                                   chunk_Aj + 1) * chunk_size_Aj]
                    pAj = np.empty((len(chunk_index_Aj), n))
                    for i in range(len(chunk_index_Aj)):
                        index = chunk_index_Aj[i]
                        wi = self._build_wi(index, self.bws_history[iter_i, j],
                                            self.taus_history[iter_i, j])
                        xw = Xj * wi
                        pAj[i, :] = Xj[index] / np.sum(xw * Xj) * xw
                    pR[chunk_index_Aj, :, j] = pAj.dot(pRj_old)
                err = pRj_old - pR[:, :, j]

        for j in range(k):
            CCT[:, j] += ((pR[:, :, j] / self.X[:, j].reshape(-1, 1)) ** 2).sum(
                axis=1)
        for i in range(len(chunk_index)):
            ENP_j += pR[chunk_index[i], i, :]

        return ENP_j, CCT,

    def fit(self, n_chunks=1):
        """
        Compute MGTWR inference by chunk to reduce memory footprint.
        """
        self.n_chunks = n_chunks
        pre = np.sum(self.X * self.betas, axis=1).reshape(-1, 1)
        result = map(self._chunk_compute, (range(n_chunks)))
        result_list = list(zip(*result))
        ENP_j = np.sum(np.array(result_list[0]), axis=0)
        CCT = np.sum(np.array(result_list[1]), axis=0)
        return MGTWRResults(
            self.coords, self.t, self.X, self.y, self.bws, self.taus, self.kernel, self.fixed, self.bw_ts,
            self.bws_history, self.taus_history, self.betas, pre, ENP_j, CCT)



import numpy as np
from typing import Union
import pandas as pd
from .diagnosis import get_AICc, get_AIC, get_BIC, get_CV
from .obj import BaseModel
from scipy.spatial.distance import pdist
from .model import GWR, GTWR
from .function import golden_section, surface_to_plane, print_time, twostep_golden_section, multi_bw, multi_bws

getDiag = {'AICc': get_AICc, 'AIC': get_AIC, 'BIC': get_BIC, 'CV': get_CV}

delta = 0.38197


class SearchGWRParameter(BaseModel):
    """
    Select bandwidth for GWR model

    Parameters
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates of
                    observations

    y             : array-like
                    n*1, dependent variable

    X             : array-like
                    n*k, independent variable, excluding the constant

    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'

    fixed         : boolean
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)

    constant      : boolean
                    True to include intercept (default) in model and False to exclude
                    intercept.

    Examples
    --------
    import numpy as np
    from mgtwr.sel import SearchGWRParameter
    np.random.seed(1)
    u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
    v = np.array([((i-1) % 144) // 12 for i in range(1, 1729)]).reshape(-1, 1)
    t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
    x1 = np.random.uniform(0, 1, (1728, 1))
    x2 = np.random.uniform(0, 1, (1728, 1))
    epsilon = np.random.randn(1728, 1)
    beta0 = 5
    beta1 = 3 + (u + v + t)/6
    beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
    y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
    coords = np.hstack([u, v])
    X = np.hstack([x1, x2])
    sel = SearchGWRParameter(coords, X, y, kernel='gaussian', fixed=True)
    bw = sel.search(bw_max=40, verbose=True)
    2.0
    """

    def __init__(
            self,
            coords: Union[np.ndarray, pd.DataFrame],
            X: Union[np.ndarray, pd.DataFrame],
            y: Union[np.ndarray, pd.DataFrame],
            kernel: str = 'exponential',
            fixed: bool = False,
            constant: bool = True,
            convert: bool = False,
            thread: int = 1
    ):

        super(SearchGWRParameter, self).__init__(X, y, kernel, fixed, constant)
        if isinstance(coords, pd.DataFrame):
            coords = coords.values
        self.coords = coords
        if convert:
            longitude = coords[:, 0]
            latitude = coords[:, 1]
            longitude, latitude = surface_to_plane(longitude, latitude)
            self.coords = np.hstack([longitude, latitude])
        self.int_score = not self.fixed
        self.thread = thread

    @print_time
    def search(self,
               criterion: str = 'AICc',
               bw_min: float = None,
               bw_max: float = None,
               tol: float = 1.0e-6,
               bw_decimal: int = 0,
               max_iter: int = 200,
               verbose: bool = True,
               time_cost: bool = False
               ):
        """
        Method to select one unique bandwidth for a GWR model.

        Parameters
        ----------
        criterion      : string
                         bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
        bw_min         : float
                         min value used in bandwidth search
        bw_max         : float
                         max value used in bandwidth search
        tol            : float
                         tolerance used to determine convergence
        max_iter       : integer
                         max iterations if no convergence to tol

        bw_decimal      : scalar
                         The number of bandwidth's decimal places saved during the search

        verbose        : bool
                         If true, bandwidth searching history is printed out; default is False.
        time_cost      : bool
                         If true, print run time
        """

        def gwr_func(x):
            return getDiag[criterion](GWR(
                self.coords, self.X, self.y, x, kernel=self.kernel,
                fixed=self.fixed, constant=False, thread=self.thread).cal_aic())

        bw_min, bw_max = self._init_section(bw_min, bw_max)
        bw = golden_section(bw_min, bw_max, delta, bw_decimal, gwr_func, tol, max_iter, verbose)
        return bw

    def _init_section(self, bw_min, bw_max):
        if bw_min is not None and bw_max is not None:
            return bw_min, bw_max

        if len(self.X) > 0:
            n_glob = self.X.shape[1]
        else:
            n_glob = 0
        if self.constant:
            n_vars = n_glob + 1
        else:
            n_vars = n_glob
        n = np.array(self.coords).shape[0]

        if self.int_score:
            a = 40 + 2 * n_vars
            c = n
        else:
            try:
                coords = np.unique(self.coords, axis=0)
                sq_dists = pdist(coords)
                a = np.min(sq_dists) / 2.0
                c = np.max(sq_dists)
            except MemoryError:
                # Note that the value obtained in this way is not the maximum distance of all points,
                # but the upper bound of the search has little effect on the results of the model
                coords = sorted(self.coords, key=lambda x: x[0] ** 2 + x[1] ** 2)
                a = pdist(coords[:2])[0]
                c = pdist([coords[0], coords[-1]])[0]
        if bw_min is None:
            bw_min = a
        if bw_max is None:
            bw_max = c

        return bw_min, bw_max


class SearchMGWRParameter(BaseModel):

    def __init__(
            self,
            coords: Union[np.ndarray, pd.DataFrame],
            X: Union[np.ndarray, pd.DataFrame],
            y: Union[np.ndarray, pd.DataFrame],
            kernel: str = 'exponential',
            fixed: bool = False,
            constant: bool = True,
            convert: bool = False,
            thread: int = 1
    ):

        super(SearchMGWRParameter, self).__init__(X, y, kernel, fixed, constant)
        if isinstance(coords, pd.DataFrame):
            coords = coords.values
        self.coords = coords
        if convert:
            longitude = coords[:, 0]
            latitude = coords[:, 1]
            longitude, latitude = surface_to_plane(longitude, latitude)
            self.coords = np.hstack([longitude, latitude])
        self.int_score = not self.fixed
        self.thread = thread
        self.criterion = None
        self.bws = None
        self.tol = None
        self.bw_decimal = None

    @print_time
    def search(
            self,
            criterion: str = 'AICc',
            bw_min: float = None,
            bw_max: float = None,
            tol: float = 1.0e-6,
            bw_decimal: int = 1,
            init_bw: float = None,
            multi_bw_min: list = None,
            multi_bw_max: list = None,
            tol_multi: float = 1.0e-5,
            bws_same_times: int = 5,
            verbose: bool = False,
            rss_score: bool = False,
            time_cost: bool = False
            ):
        """
        Method to select one unique bandwidth and Spatio-temporal scale for a gtwr model or a
        bandwidth vector and Spatio-temporal scale vector for a mgwr model.

        Parameters
        ----------
        criterion      : string
                         bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
        bw_min         : float
                         min value used in bandwidth search
        bw_max         : float
                         max value used in bandwidth search
        multi_bw_min   : list
                         min values used for each covariate in mgwr bandwidth search.
                         Must be either a single value or have one value for
                         each covariate including the intercept
        multi_bw_max   : list
                         max values used for each covariate in mgwr bandwidth
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        tol            : float
                         tolerance used to determine convergence
        bw_decimal     : int
                        The number of bw decimal places reserved
        init_bw        : float
                         None (default) to initialize MGTWR with a bandwidth
                         derived from GTWR. Otherwise this option will choose the
                         bandwidth to initialize MGWR with.
        tol_multi      : convergence tolerance for the multiple bandwidth
                         back fitting algorithm; a larger tolerance may stop the
                         algorithm faster though it may result in a less optimal
                         model
        bws_same_times : If bandwidths keep the same between iterations for
                         bws_same_times (default 5) in backfitting, then use the
                         current set of bandwidths as final bandwidths.
        rss_score      : True to use the residual sum of squares to evaluate
                         each iteration of the multiple bandwidth back fitting
                         routine and False to use a smooth function; default is
                         False
        verbose        : Boolean
                         If true, bandwidth searching history is printed out; default is False.
        time_cost      : bool
                        If true, print run time
        """
        self.criterion = criterion
        self.tol = tol
        self.bw_decimal = bw_decimal
        if multi_bw_min is not None:
            if len(multi_bw_min) == self.k:
                multi_bw_min = multi_bw_min
            elif len(multi_bw_min) == 1:
                multi_bw_min = multi_bw_min * self.k
            else:
                raise AttributeError(
                    "multi_bw_min must be either a list containing"
                    " a single entry or a list containing an entry for each of k"
                    " covariates including the intercept")
        else:
            a = self._init_section(bw_min, bw_max)[0]
            multi_bw_min = [a] * self.k

        if multi_bw_max is not None:
            if len(multi_bw_max) == self.k:
                multi_bw_max = multi_bw_max
            elif len(multi_bw_max) == 1:
                multi_bw_max = multi_bw_max * self.k
            else:
                raise AttributeError(
                    "multi_bw_max must be either a list containing"
                    " a single entry or a list containing an entry for each of k"
                    " covariates including the intercept")
        else:
            c = self._init_section(bw_min, bw_max)[1]
            multi_bw_max = [c] * self.k

        self.bws = multi_bw(init_bw, self.X, self.y, self.n, self.k, tol_multi,
                            rss_score, self.gwr_func, self.bw_func, self.sel_func, multi_bw_min, multi_bw_max,
                            bws_same_times, verbose=verbose)
        return self.bws

    def gwr_func(self, X, y, bw):
        res = GWR(self.coords, X, y, bw, kernel=self.kernel,
                  fixed=self.fixed, constant=False, thread=self.thread).cal_multi()
        return res

    def bw_func(self, X, y):
        selector = SearchGWRParameter(self.coords, X, y, kernel=self.kernel, fixed=self.fixed,
                                      constant=False, thread=self.thread)
        return selector

    def sel_func(self, bw_func, bw_min=None, bw_max=None):
        return bw_func.search(criterion=self.criterion, bw_min=bw_min, bw_max=bw_max,
                              tol=self.tol, bw_decimal=self.bw_decimal, verbose=False)

    def _init_section(self, bw_min, bw_max):

        a = bw_min if bw_min is not None else 0
        if bw_max is not None:
            c = bw_max
        else:
            c = max(np.max(self.coords[:, 0]) - np.min(self.coords[:, 0]),
                    np.max(self.coords[:, 1]) - np.min(self.coords[:, 1]))
        return a, c


class SearchGTWRParameter(BaseModel):
    """
    Select bandwidth for GTWR model

    Parameters
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates of
                    observations

    t             : array-like
                    n*1, time location

    y             : array-like
                    n*1, dependent variable

    X             : array-like
                    n*k, independent variable, excluding the constant

    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'

    fixed         : boolean
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)

    constant      : boolean
                    True to include intercept (default) in model and False to exclude
                    intercept.

    Examples
    --------
    import numpy as np
    from mgtwr.sel import SearchGTWRParameter
    np.random.seed(1)
    u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
    v = np.array([((i-1) % 144)//12 for i in range(1, 1729)]).reshape(-1, 1)
    t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
    x1 = np.random.uniform(0, 1, (1728, 1))
    x2 = np.random.uniform(0, 1, (1728, 1))
    epsilon = np.random.randn(1728, 1)
    beta0 = 5
    beta1 = 3 + (u + v + t)/6
    beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
    y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
    coords = np.hstack([u, v])
    X = np.hstack([x1, x2])
    sel = SearchGTWRParameter(coords, t, X, y, kernel='gaussian', fixed=True)
    bw, tau = sel.search(tau_max=20, verbose=True)
    0.9, 1.5
    """

    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            kernel: str = 'exponential',
            fixed: bool = False,
            constant: bool = True,
            convert: bool = False,
            thread: int = 1
    ):

        super(SearchGTWRParameter, self).__init__(X, y, kernel, fixed, constant)
        if isinstance(coords, pd.DataFrame):
            coords = coords.values
        self.coords = coords
        if convert:
            longitude = coords[:, 0]
            latitude = coords[:, 1]
            longitude, latitude = surface_to_plane(longitude, latitude)
            self.coords = np.hstack([longitude, latitude])
        self.t = t
        self.int_score = not self.fixed
        self.thread = thread

    @print_time
    def search(
            self,
            criterion: str = 'AICc',
            bw_min: float = None,
            bw_max: float = None,
            tau_min: float = None,
            tau_max: float = None,
            tol: float = 1.0e-6,
            bw_decimal: int = 1,
            tau_decimal: int = 1,
            max_iter: int = 200,
            verbose: bool = False,
            time_cost: bool = False
            ):
        """
        Method to select one unique bandwidth and Spatio-temporal scale for a GTWR model.

        Parameters
        ----------
        criterion      : string
                         bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
        bw_min         : float
                         min value used in bandwidth search
        bw_max         : float
                         max value used in bandwidth search
        tau_min        : float
                         min value used in spatio-temporal scale search
        tau_max        : float
                         max value used in spatio-temporal scale search
        tol            : float
                         tolerance used to determine convergence
        max_iter       : integer
                         max iterations if no convergence to tol
        bw_decimal      : scalar
                         The number of bandwidth's decimal places saved during the search
        tau_decimal     : scalar
                         The number of Spatio-temporal decimal places saved during the search
        verbose        : Boolean
                         If true, bandwidth searching history is printed out; default is False.
        time_cost      : bool
                        If true, print run time
        """

        def gtwr_func(x, y):
            return getDiag[criterion](GTWR(
                self.coords, self.t, self.X, self.y, x, y, kernel=self.kernel,
                fixed=self.fixed, constant=False, thread=self.thread).cal_aic())

        bw_min, bw_max, tau_min, tau_max = self._init_section(bw_min, bw_max, tau_min, tau_max)
        bw, tau = twostep_golden_section(bw_min, bw_max, tau_min, tau_max, delta, gtwr_func, tol, max_iter, bw_decimal,
                                         tau_decimal, verbose)

        return bw, tau

    def _init_section(self, bw_min, bw_max, tau_min, tau_max):
        if (bw_min is not None) and (bw_max is not None) and (tau_min is not None) and (tau_max is not None):
            return bw_min, bw_max, tau_min, tau_max
        if len(self.X) > 0:
            n_glob = self.X.shape[1]
        else:
            n_glob = 0
        if self.constant:
            n_vars = n_glob + 1
        else:
            n_vars = n_glob
        n = np.array(self.coords).shape[0]

        if self.int_score:
            a = 40 + 2 * n_vars
            c = n
        else:
            try:
                coords = np.unique(self.coords, axis=0)
                sq_dists = pdist(coords)
                a = np.min(sq_dists) / 2.0
                c = np.max(sq_dists)
            except MemoryError:
                # Note that the value obtained in this way is not the maximum distance of all points,
                # but the upper bound of the search has little effect on the results of the model
                coords = sorted(self.coords, key=lambda x: x[0] ** 2 + x[1] ** 2)
                a = pdist(coords[:2])[0]
                c = pdist([coords[0], coords[-1]])[0]

        if bw_min is None:
            bw_min = a
        if bw_max is None:
            bw_max = c

        if tau_min is None:
            tau_min = 0
        if tau_max is None:
            tau_max = 4
        return bw_min, bw_max, tau_min, tau_max


class SearchMGTWRParameter(BaseModel):
    """
    Select bandwidth for MGTWR model

    Parameters
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates of
                    observations

    t             : array-like
                    n*1, time location

    X             : array-like
                    n*k, independent variable, excluding the constant

    y             : array-like
                    n*1, dependent variable

    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'

    fixed         : bool
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)

    constant      : bool
                    True to include intercept (default) in model and False to exclude
                    intercept.

    Examples
    --------
    import numpy as np
    from mgtwr.sel import SearchMGTWRParameter
    from mgtwr.model import MGTWR
    np.random.seed(10)
    u = np.array([(i-1) % 12 for i in range(1, 1729)]).reshape(-1, 1)
    v = np.array([((i-1) % 144)//12 for i in range(1, 1729)]).reshape(-1, 1)
    t = np.array([(i-1) // 144 for i in range(1, 1729)]).reshape(-1, 1)
    x1 = np.random.uniform(0, 1, (1728, 1))
    x2 = np.random.uniform(0, 1, (1728, 1))
    epsilon = np.random.randn(1728, 1)
    beta0 = 5
    beta1 = 3 + (u + v + t)/6
    beta2 = 3 + ((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2)) / 128
    y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
    coords = np.hstack([u, v])
    X = np.hstack([x1, x2])
    sel_multi = SearchMGTWRParameter(coords, t, X, y, kernel='gaussian', fixed=True)
    bws = sel_multi.search(multi_bw_min=[0.1], verbose=True, tol_multi=1.0e-4)
    mgtwr = MGTWR(coords, t, X, y, sel_multi, kernel='gaussian', fixed=True).fit()
    print(mgtwr.R2)
    0.9972924820674222
    """
    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            kernel: str = 'exponential',
            fixed: bool = False,
            constant: bool = True,
            convert: bool = False,
            thread: int = 1
    ):

        super(SearchMGTWRParameter, self).__init__(X, y, kernel, fixed, constant)
        if isinstance(coords, pd.DataFrame):
            coords = coords.values
        self.coords = coords
        if convert:
            longitude = coords[:, 0]
            latitude = coords[:, 1]
            longitude, latitude = surface_to_plane(longitude, latitude)
            self.coords = np.hstack([longitude, latitude])
        self.t = t
        self.int_score = not self.fixed
        self.thread = thread
        self.criterion = None
        self.bws = None
        self.tol = None
        self.bw_decimal = None
        self.tau_decimal = None

    @print_time
    def search(
            self,
            criterion: str = 'AICc',
            bw_min: float = None,
            bw_max: float = None,
            tau_min: float = None,
            tau_max: float = None,
            tol: float = 1.0e-6,
            bw_decimal: int = 1,
            tau_decimal: int = 1,
            init_bw: float = None,
            init_tau: float = None,
            multi_bw_min: list = None,
            multi_bw_max: list = None,
            multi_tau_min: list = None,
            multi_tau_max: list = None,
            tol_multi: float = 1.0e-5,
            verbose: bool = False,
            rss_score: bool = False,
            time_cost: bool = False
            ):
        """
        Method to select one unique bandwidth and Spatio-temporal scale for a gtwr model or a
        bandwidth vector and Spatio-temporal scale vector for a mtgwr model.

        Parameters
        ----------
        criterion      : string
                         bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
        bw_min         : float
                         min value used in bandwidth search
        bw_max         : float
                         max value used in bandwidth search
        tau_min        : float
                         min value used in spatio-temporal scale search
        tau_max        : float
                         max value used in spatio-temporal scale search
        multi_bw_min   : list
                         min values used for each covariate in mgwr bandwidth search.
                         Must be either a single value or have one value for
                         each covariate including the intercept
        multi_bw_max   : list
                         max values used for each covariate in mgwr bandwidth
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        multi_tau_min  : list
                         min values used for each covariate in mgtwr spatio-temporal scale
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        multi_tau_max  : max values used for each covariate in mgtwr spatio-temporal scale
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        tol            : float
                         tolerance used to determine convergence
        bw_decimal     : int
                        The number of bw decimal places reserved
        tau_decimal    : int
                        The number of tau decimal places reserved
        init_bw        : float
                         None (default) to initialize MGTWR with a bandwidth
                         derived from GTWR. Otherwise this option will choose the
                         bandwidth to initialize MGWR with.
        init_tau       : float
                         None (default) to initialize MGTWR with a spatio-temporal scale
                         derived from GTWR. Otherwise this option will choose the
                         spatio-temporal scale to initialize MGWR with.
        tol_multi      : convergence tolerance for the multiple bandwidth
                         back fitting algorithm; a larger tolerance may stop the
                         algorithm faster though it may result in a less optimal
                         model
        rss_score      : True to use the residual sum of squares to evaluate
                         each iteration of the multiple bandwidth back fitting
                         routine and False to use a smooth function; default is
                         False
        verbose        : Boolean
                         If true, bandwidth searching history is printed out; default is False.
        time_cost      : bool
                        If true, print run time
        """
        self.criterion = criterion
        self.tol = tol
        self.bw_decimal = bw_decimal
        self.tau_decimal = tau_decimal
        if multi_bw_min is not None:
            if len(multi_bw_min) == self.k:
                multi_bw_min = multi_bw_min
            elif len(multi_bw_min) == 1:
                multi_bw_min = multi_bw_min * self.k
            else:
                raise AttributeError(
                    "multi_bw_min must be either a list containing"
                    " a single entry or a list containing an entry for each of k"
                    " covariates including the intercept")
        else:
            a = self._init_section(bw_min, bw_max, tau_min, tau_max)[0]
            multi_bw_min = [a] * self.k

        if multi_bw_max is not None:
            if len(multi_bw_max) == self.k:
                multi_bw_max = multi_bw_max
            elif len(multi_bw_max) == 1:
                multi_bw_max = multi_bw_max * self.k
            else:
                raise AttributeError(
                    "multi_bw_max must be either a list containing"
                    " a single entry or a list containing an entry for each of k"
                    " covariates including the intercept")
        else:
            c = self._init_section(bw_min, bw_max, tau_min, tau_max)[1]
            multi_bw_max = [c] * self.k

        if multi_tau_min is not None:
            if len(multi_tau_min) == self.k:
                multi_tau_min = multi_tau_min
            elif len(multi_tau_min) == 1:
                multi_tau_min = multi_tau_min * self.k
            else:
                raise AttributeError(
                    "multi_tau_min must be either a list containing"
                    " a single entry or a list containing an entry for each of k"
                    " variates including the intercept")
        else:
            A = self._init_section(bw_min, bw_max, tau_min, tau_max)[2]
            multi_tau_min = [A] * self.k

        if multi_tau_max is not None:
            if len(multi_tau_max) == self.k:
                multi_tau_max = multi_tau_max
            elif len(multi_tau_max) == 1:
                multi_tau_max = multi_tau_max * self.k
            else:
                raise AttributeError(
                    "multi_tau_max must be either a list containing"
                    " a single entry or a list containing an entry for each of k"
                    " variates including the intercept")
        else:
            C = self._init_section(bw_min, bw_max, tau_min, tau_max)[3]
            multi_tau_max = [C] * self.k

        self.bws = multi_bws(init_bw, init_tau, self.X, self.y, self.n, self.k, tol_multi,
                             rss_score, self.gtwr_func, self.bw_func, self.sel_func, multi_bw_min, multi_bw_max,
                             multi_tau_min, multi_tau_max, verbose=verbose)
        return self.bws

    def gtwr_func(self, X, y, bw, tau):
        return GTWR(self.coords, self.t, X, y, bw, tau, kernel=self.kernel,
                    fixed=self.fixed, constant=False, thread=self.thread).cal_multi()

    def bw_func(self, X, y):
        selector = SearchGTWRParameter(self.coords, self.t, X, y, kernel=self.kernel, fixed=self.fixed,
                                       constant=False, thread=self.thread)
        return selector

    def sel_func(self, bw_func, bw_min=None, bw_max=None, tau_min=None, tau_max=None):
        return bw_func.search(criterion=self.criterion, bw_min=bw_min, bw_max=bw_max, tau_min=tau_min, tau_max=tau_max,
                              tol=self.tol, bw_decimal=self.bw_decimal, tau_decimal=self.tau_decimal, verbose=False)

    def _init_section(self, bw_min, bw_max, tau_min, tau_max):

        a = bw_min if bw_min is not None else 0
        if bw_max is not None:
            c = bw_max
        else:
            c = max(np.max(self.coords[:, 0]) - np.min(self.coords[:, 0]),
                    np.max(self.coords[:, 1]) - np.min(self.coords[:, 1]))

        A = tau_min if tau_min is not None else 0
        C = tau_max if tau_max is not None else 4

        return a, c, A, C



import numpy as np
from scipy.spatial.distance import cdist


class GWRKernel:

    def __init__(
            self,
            coords: np.ndarray,
            bw: float = None,
            fixed: bool = True,
            function: str = 'triangular',
            eps: float = 1.0000001):

        self.coords = coords
        self.function = function
        self.bw = bw
        self.fixed = fixed
        self.function = function
        self.eps = eps
        self.bandwidth = None
        self.kernel = None

    def cal_distance(
            self,
            i: int):
        distance = cdist([self.coords[i]], self.coords).reshape(-1)
        return distance

    def cal_kernel(
            self,
            distance
    ):

        if self.fixed:
            self.bandwidth = float(self.bw)
        else:
            self.bandwidth = np.partition(
                distance,
                int(self.bw) - 1)[int(self.bw) - 1] * self.eps  # partial sort in O(n) Time

        self.kernel = self._kernel_funcs(distance / self.bandwidth)

        if self.function == "bisquare":  # Truncate for bisquare
            self.kernel[(distance >= self.bandwidth)] = 0
        return self.kernel

    def _kernel_funcs(self, zs):
        # functions follow Anselin and Rey (2010) table 5.4
        if self.function == 'triangular':
            return 1 - zs
        elif self.function == 'uniform':
            return np.ones(zs.shape) * 0.5
        elif self.function == 'quadratic':
            return (3. / 4) * (1 - zs ** 2)
        elif self.function == 'quartic':
            return (15. / 16) * (1 - zs ** 2) ** 2
        elif self.function == 'gaussian':
            return np.exp(-0.5 * zs ** 2)
        elif self.function == 'bisquare':
            return (1 - zs ** 2) ** 2
        elif self.function == 'exponential':
            return np.exp(-zs)
        else:
            print('Unsupported kernel function', self.function)


class GTWRKernel(GWRKernel):

    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            bw: float = None,
            tau: float = None,
            fixed: bool = True,
            function: str = 'triangular',
            eps: float = 1.0000001):

        super(GTWRKernel, self).__init__(coords, bw, fixed=fixed, function=function, eps=eps)

        self.t = t
        self.tau = tau
        self.coords_new = None

    def cal_distance(
            self,
            i: int):

        if self.tau == 0:
            self.coords_new = self.coords
        else:
            self.coords_new = np.hstack([self.coords, (np.sqrt(self.tau) * self.t)])
        distance = cdist([self.coords_new[i]], self.coords_new).reshape(-1)
        return distance



import numpy as np
from scipy import linalg
import time
from typing import Callable
from copy import deepcopy


def print_time(func: Callable):
    def inner(*args, **kwargs):
        start = time.time()
        res = func(*args, **kwargs)
        end = time.time()
        m, s = divmod(end - start, 60)
        h, m = divmod(m, 60)
        if 'time_cost' in kwargs and kwargs['time_cost']:
            print("time cost: %d:%02d:%s" % (h, m, round(s, 3)))
        return res
    return inner


def _compute_betas_gwr(y, x, wi):
    """
    compute MLE coefficients using iwls routine

    Methods: p189, Iteratively (Re)weighted Least Squares (IWLS),
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.
    """
    xt = (x * wi).T
    xtx = np.dot(xt, x)
    xtx_inv_xt = linalg.solve(xtx, xt)
    betas = np.dot(xtx_inv_xt, y)
    return betas, xtx_inv_xt


def surface_to_plane(
        longitude: np.ndarray,
        latitude: np.ndarray,
        central_longitude: int = 114
):

    r"""
    base on Gauss-Kruger projection

    equatorial radius: a = 6378136.49m

    polar radius: b = 6356755m

    so that

    first eccentricity :math:`e = \sqrt{a^2-b^2}/a`

    second eccentricity :math:`e' = \sqrt{a^2-b^2}/b`

    so that

    .. math::
        \begin{aligned}
            Y_{b0}=a^2B\beta_0/b +
            sin(B)\left(\beta_2cos(B)+\beta_4cos^3(B)+\beta_6cos^5(B)+\beta_8cos^7(B)\right)
        \end{aligned}
    where B is the latitude converted from degrees to radians and

    .. math::
        \begin{aligned}
            \beta_0 &= 1-\frac{3}{4}e'^2+\frac{45}{64}e'^4-\frac{175}{256}e'^6+
                       \frac{11025}{16384}e'^8 \\
            \beta_2 &= \beta_0 - 1 \\
            \beta_4 &= \frac{15}{32}e'^4-\frac{175}{384}e'^6+\frac{3675}{8192}e'^8 \\
            \beta_6 &= -\frac{35}{96}e'^6 + \frac{735}{2048}e'^8 \\
            \beta_8 &= \frac{315}{1024}e'^8 \\
        \end{aligned}

    so that the Y-axis is

    .. math::
        \begin{aligned}
            Y &= Y_{b0}+\frac{1}{2}Ntan(B)m^2+\frac{1}{24}\left(5-tan^2(B)+9\eta^2+4\eta^4
                \right)Ntan(B)m^4 \\
              &+ \frac{1}{720}\left(61-58tan^2(B)\right)Ntan(B)m^6
        \end{aligned}
    where L is the longitude subtracts the central longitude converted to radians and

    .. math::
        \begin{aligned}
            N &= a/\sqrt{1-(esin(B))^2} \\
            \eta &= e'cos(B) \\
            m &= Lcos(B) \\
        \end{aligned}
    so that the X_axis is

    .. math::
        \begin{aligned}
            X &= Nm+\frac{1}{6}\left(1-tan^2(B)+\eta^2\right)Nm^3 \\
              &+ \frac{1}{120}\left(5-18tan^2(B)+tan^4(B)+14\eta^2-58tan^2(B)\eta\right)Nm^5+500000
        \end{aligned}
    """
    a = 6378136.49
    b = 6356755

    e1 = np.sqrt(a ** 2 - b ** 2) / a
    e2 = np.sqrt(a ** 2 - b ** 2) / b
    beta0 = 1 - (3 / 4) * e2 ** 2 + (45 / 64) * e2 ** 4 - (175 / 256) * e2 ** 6 \
        + (11025 / 16384) * e2 ** 8
    beta2 = beta0 - 1
    beta4 = (15 / 32) * e2 ** 4 - (175 / 384) * e2 ** 6 + (3675 / 8192) * e2 ** 8
    beta6 = -(35 / 96) * e2 ** 6 + (735 / 2048) * e2 ** 8
    beta8 = (315 / 1024) * e2 ** 8

    L = np.radians(longitude - central_longitude)
    B = np.radians(latitude)
    cosB = np.cos(B)
    sinB = np.sin(B)
    tanB = np.tan(B)
    N = a / np.sqrt(1 - (e1 * sinB) ** 2)
    eta = e2 * cosB
    m = L * cosB
    Yb0 = a ** 2 * B * beta0 / b + sinB * \
        (beta2 * cosB + beta4 * cosB ** 3 + beta6 * cosB ** 5 + beta8 * cosB ** 7)
    Y = Yb0 + (1 / 2) * N * tanB * m ** 2 + (1 / 24) * (5 - tanB ** 2 + 9 * eta ** 2 + 4 * eta ** 4) * N * tanB * \
        m ** 4 + (1 / 720) * (61 - 58 * tanB ** 2) * N * tanB * m ** 6
    X = N * m + (1 / 6) * (1 - tanB ** 2 + eta ** 2) * N * m ** 3 + \
        (1 / 120) * (5 - 18 * tanB ** 2 + tanB ** 4 + 14 * eta ** 2 - 58 * tanB ** 2 * eta) * N * m ** 5 + 500000
    X = X.reshape(-1, 1)
    Y = Y.reshape(-1, 1)
    return X, Y


def golden_section(a, c, delta, decimal, function, tol, max_iter, verbose=False):
    b = a + delta * np.abs(c - a)
    d = c - delta * np.abs(c - a)
    diff = 1.0e9
    iter_num = 0
    score_dict = {}
    opt_val = None
    while np.abs(diff) > tol and iter_num < max_iter:
        iter_num += 1
        b = np.round(b, decimal)
        d = np.round(d, decimal)

        if b in score_dict:
            score_b = score_dict[b]
        else:
            score_b = function(b)
            score_dict[b] = score_b

        if d in score_dict:
            score_d = score_dict[d]
        else:
            score_d = function(d)
            score_dict[d] = score_d

        if score_b <= score_d:
            opt_val = b
            opt_score = score_b
            c = d
            d = b
            b = a + delta * np.abs(c - a)

        else:
            opt_val = d
            opt_score = score_d
            a = b
            b = d
            d = c - delta * np.abs(c - a)

        opt_val = np.round(opt_val, decimal)
        diff = score_b - score_d
        if verbose:
            print('bw:', opt_val, ', score:', np.round(opt_score, 2))

    return opt_val


def onestep_golden_section(A, C, x, delta, tau_decimal, function, tol):
    iter_num = 0
    score_dict = {}
    diff = 1e9
    opt_score = None
    opt_tau = None
    B = A + delta * np.abs(C - A)
    D = C - delta * np.abs(C - A)
    while np.abs(diff) > tol and iter_num < 200:
        iter_num += 1
        B = np.round(B, tau_decimal)
        D = np.round(D, tau_decimal)
        if B in score_dict:
            score_B = score_dict[B]
        else:
            score_B = function(x, B)
            score_dict[B] = score_B

        if D in score_dict:
            score_D = score_dict[D]
        else:
            score_D = function(x, D)
            score_dict[D] = score_D
        if score_B <= score_D:
            opt_score = score_B
            opt_tau = B
            C = D
            D = B
            B = A + delta * np.abs(C - A)
        else:
            opt_score = score_D
            opt_tau = D
            A = B
            B = D
            D = C - delta * np.abs(C - A)
        diff = score_B - score_D
    return opt_tau, opt_score


def twostep_golden_section(
        a, c, A, C, delta, function,
        tol, max_iter, bw_decimal, tau_decimal, verbose=False):
    b = a + delta * np.abs(c - a)
    d = c - delta * np.abs(c - a)
    opt_bw = None
    opt_tau = None
    diff = 1e9
    score_dict = {}
    iter_num = 0
    while np.abs(diff) > tol and iter_num < max_iter:
        iter_num += 1
        b = np.round(b, bw_decimal)
        d = np.round(d, bw_decimal)
        if b in score_dict:
            tau_b, score_b = score_dict[b]
        else:
            tau_b, score_b = onestep_golden_section(A, C, b, delta, tau_decimal, function, tol)
            score_dict[b] = [tau_b, score_b]
        if d in score_dict:
            tau_d, score_d = score_dict[d]
        else:
            tau_d, score_d = onestep_golden_section(A, C, d, delta, tau_decimal, function, tol)
            score_dict[d] = [tau_d, score_d]

        if score_b <= score_d:
            opt_score = score_b
            opt_bw = b
            opt_tau = tau_b
            c = d
            d = b
            b = a + delta * np.abs(c - a)
        else:
            opt_score = score_d
            opt_bw = d
            opt_tau = tau_d
            a = b
            b = d
            d = c - delta * np.abs(c - a)
        diff = score_b - score_d
        if verbose:
            print('bw: ', opt_bw, ', tau: ', opt_tau, ', score: ', opt_score)
    return opt_bw, opt_tau


def multi_bw(init, X, y, n, k, tol, rss_score, gwr_func,
             bw_func, sel_func, multi_bw_min, multi_bw_max, bws_same_times,
             verbose=False):
    """
    Multiscale GWR bandwidth search procedure using iterative GAM backfitting
    """
    if init is None:
        bw = sel_func(bw_func(X, y))
        optim_model = gwr_func(X, y, bw)
    else:
        bw = init
        optim_model = gwr_func(X, y, init)
    bw_gwr = bw
    err = optim_model.reside
    betas = optim_model.betas
    XB = np.multiply(betas, X)
    rss = np.sum(err ** 2) if rss_score else None
    scores = []
    BWs = []
    bw_stable_counter = 0
    bws = np.empty(k)
    Betas = None

    for iters in range(1, 201):
        new_XB = np.zeros_like(X)
        Betas = np.zeros_like(X)

        for j in range(k):
            temp_y = XB[:, j].reshape((-1, 1))
            temp_y = temp_y + err
            temp_X = X[:, j].reshape((-1, 1))
            bw_class = bw_func(temp_X, temp_y)

            if bw_stable_counter >= bws_same_times:
                # If in backfitting, all bws not changing in bws_same_times (default 5) iterations
                bw = bws[j]
            else:
                bw = sel_func(bw_class, multi_bw_min[j], multi_bw_max[j])

            optim_model = gwr_func(temp_X, temp_y, bw)
            err = optim_model.reside
            betas = optim_model.betas
            new_XB[:, j] = optim_model.pre.reshape(-1)
            Betas[:, j] = betas.reshape(-1)
            bws[j] = bw

        # If bws remain the same as from previous iteration
        if (iters > 1) and np.all(BWs[-1] == bws):
            bw_stable_counter += 1
        else:
            bw_stable_counter = 0

        num = np.sum((new_XB - XB) ** 2) / n
        den = np.sum(np.sum(new_XB, axis=1) ** 2)
        score = (num / den) ** 0.5
        XB = new_XB

        if rss_score:
            predy = np.sum(np.multiply(betas, X), axis=1).reshape((-1, 1))
            new_rss = np.sum((y - predy) ** 2)
            score = np.abs((new_rss - rss) / new_rss)
            rss = new_rss
        scores.append(deepcopy(score))
        delta = score
        BWs.append(deepcopy(bws))

        if verbose:
            print("Current iteration:", iters, ",SOC:", np.round(score, 7))
            print("Bandwidths:", ', '.join([str(bw) for bw in bws]))

        if delta < tol:
            break

    opt_bw = BWs[-1]
    return opt_bw, np.array(BWs), np.array(scores), Betas, err, bw_gwr


def multi_bws(init_bw, init_tau, X, y, n, k, tol, rss_score,
              gtwr_func, bw_func, sel_func, multi_bw_min, multi_bw_max,
              multi_tau_min, multi_tau_max, verbose=False):
    """
    Multiscale GTWR bandwidth search procedure using iterative GAM back fitting
    """
    if (init_bw is None) or (init_tau is None):
        bw, tau = sel_func(bw_func(X, y))
    else:
        bw, tau = init_bw, init_tau
    opt_model = gtwr_func(X, y, bw, tau)
    bw_gtwr = bw
    tau_gtwr = tau
    err = opt_model.reside
    betas = opt_model.betas

    XB = np.multiply(betas, X)
    rss = np.sum(err ** 2) if rss_score else None
    scores = []
    bws = np.empty(k)
    taus = np.empty(k)
    BWs = []
    Taus = []
    Betas = None

    for iter_num in range(1, 201):
        new_XB = np.zeros_like(X)
        Betas = np.zeros_like(X)

        for j in range(k):
            temp_y = XB[:, j].reshape((-1, 1))
            temp_y = temp_y + err
            temp_X = X[:, j].reshape((-1, 1))
            bw_class = bw_func(temp_X, temp_y)

            bw, tau = sel_func(bw_class, multi_bw_min[j], multi_bw_max[j],
                               multi_tau_min[j], multi_tau_max[j])

            opt_model = gtwr_func(temp_X, temp_y, bw, tau)
            err = opt_model.reside
            betas = opt_model.betas
            new_XB[:, j] = (betas * temp_X).reshape(-1)
            Betas[:, j] = betas.reshape(-1)
            bws[j] = bw
            taus[j] = tau

        num = np.sum((new_XB - XB) ** 2) / n
        den = np.sum(np.sum(new_XB, axis=1) ** 2)
        score = (num / den) ** 0.5
        XB = new_XB

        if rss_score:
            predy = np.sum(np.multiply(betas, X), axis=1).reshape((-1, 1))
            new_rss = np.sum((y - predy) ** 2)
            score = np.abs((new_rss - rss) / new_rss)
            rss = new_rss
        scores.append(deepcopy(score))
        delta = score
        BWs.append(deepcopy(bws))
        Taus.append(deepcopy(taus))

        if verbose:
            print("Current iteration:", iter_num, ",SOC:", np.round(score, 7))
            print("Bandwidths:", ', '.join([str(bw) for bw in bws]))
            print("taus:", ','.join([str(tau) for tau in taus]))

        if delta < tol:
            break
    opt_bws = BWs[-1]
    opt_tau = Taus[-1]
    return (opt_bws, opt_tau, np.array(BWs), np.array(Taus), np.array(scores),
            Betas, err, bw_gtwr, tau_gtwr)



import setuptools

setuptools.setup(
    version="2.0.4",
    long_description="To fit geographically weighted model, "
                     "multiscale geographically weighted regression model, "
                     "geographically and temporally weighted regression model and "
                     "multiscale geographically and temporally weighted regression model.",
    author="Kun Sun",
    author_email="849024477@qq.com",
    packages=['mgtwr'],
    url="https://github.com/sunkun1997/mgtwr",
)



import numpy as np
import pandas as pd
from typing import Union


class CalAicObj:

    def __init__(self, tr_S, llf, aa, n):
        self.tr_S = tr_S
        self.llf = llf
        self.aa = aa
        self.n = n


class CalMultiObj:

    def __init__(self, betas, pre, reside):
        self.betas = betas
        self.pre = pre
        self.reside = reside


class BaseModel:
    """
    Is the parent class of most models
    """
    def __init__(
            self,
            X: Union[np.ndarray, pd.DataFrame, pd.Series],
            y: Union[np.ndarray, pd.DataFrame, pd.Series],
            kernel: str,
            fixed: bool,
            constant: bool,
    ):
        self.X = X.values if isinstance(X, (pd.DataFrame, pd.Series)) else X
        self.y = y.values if isinstance(y, (pd.DataFrame, pd.Series)) else y
        if len(y.shape) > 1 and y.shape[1] != 1:
            raise ValueError('Label should be one-dimensional arrays')
        if len(y.shape) == 1:
            self.y = self.y.reshape(-1, 1)
        self.kernel = kernel
        self.fixed = fixed
        self.constant = constant
        self.n = X.shape[0]
        if self.constant:
            if len(self.X.shape) == 1 and np.all(self.X == 1):
                raise ValueError("You've already passed in a constant sequence, use constant=False instead")
            for j in range(self.X.shape[1]):
                if np.all(self.X[:, j] == 1):
                    raise ValueError("You've already passed in a constant sequence, use constant=False instead")
            self.X = np.hstack([np.ones((self.n, 1)), X])
        self.k = self.X.shape[1]


class Results(BaseModel):
    """
    Is the result parent class of all models
    """

    def __init__(
            self,
            X: Union[np.ndarray, pd.DataFrame],
            y: Union[np.ndarray, pd.Series],
            kernel: str,
            fixed: bool,
            influ: np.ndarray,
            reside,
            predict_value: np.ndarray,
            betas: np.ndarray,
            tr_STS: float
    ):
        super(Results, self).__init__(X, y, kernel, fixed, constant=False)
        self.influ = influ
        self.reside = reside
        self.predict_value = predict_value
        self.betas = betas
        self.tr_S = np.sum(influ)
        self.ENP = self.tr_S
        self.tr_STS = tr_STS
        self.TSS = np.sum((y - np.mean(y)) ** 2)
        self.RSS = np.sum(reside ** 2)
        self.sigma2 = self.RSS / (self.n - self.tr_S)
        self.std_res = self.reside / (np.sqrt(self.sigma2 * (1.0 - self.influ)))
        self.cooksD = self.std_res ** 2 * self.influ / (self.tr_S * (1.0 - self.influ))
        self.df_model = self.n - self.tr_S
        self.df_reside = self.n - 2.0 * self.tr_S + self.tr_STS
        self.R2 = 1 - self.RSS / self.TSS
        self.adj_R2 = 1 - (1 - self.R2) * (self.n - 1) / (self.n - self.ENP - 1)
        self.llf = -np.log(self.RSS) * self.n / 2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
        self.aic = -2.0 * self.llf + 2.0 * (self.tr_S + 1)
        self.aicc = self.aic + 2.0 * self.tr_S * (self.tr_S + 1.0) / (self.n - self.tr_S - 1.0)
        self.bic = -2.0 * self.llf + (self.k + 1) * np.log(self.n)


class GWRResults(Results):

    def __init__(
            self, coords, X, y, bw, kernel, fixed, influ, reside, predict_value, betas, CCT, tr_STS
    ):
        """
        betas               : array
                              n*k, estimated coefficients

        predict             : array
                              n*1, predict y values

        CCT                 : array
                              n*k, scaled variance-covariance matrix

        df_model            : integer
                              model degrees of freedom

        df_reside           : integer
                              residual degrees of freedom

        reside              : array
                              n*1, residuals of the response

        RSS                 : scalar
                              residual sum of squares

        CCT                 : array
                              n*k, scaled variance-covariance matrix

        ENP                 : scalar
                              effective number of parameters, which depends on
                              sigma2

        tr_S                : float
                              trace of S (hat) matrix

        tr_STS              : float
                              trace of STS matrix

        R2                  : float
                              R-squared for the entire model (1- RSS/TSS)

        adj_R2              : float
                              adjusted R-squared for the entire model

        aic                 : float
                              Akaike information criterion

        aicc                : float
                              corrected Akaike information criterion
                              to account for model complexity (smaller
                              bandwidths)

        bic                 : float
                              Bayesian information criterion

        sigma2              : float
                              sigma squared (residual variance) that has been
                              corrected to account for the ENP

        std_res             : array
                              n*1, standardised residuals

        bse                 : array
                              n*k, standard errors of parameters (betas)

        influ               : array
                              n*1, leading diagonal of S matrix

        CooksD              : array
                              n*1, Cook's D

        tvalues             : array
                              n*k, local t-statistics

        llf                 : scalar
                              log-likelihood of the full model; see
                              pysal.contrib.glm.family for damily-sepcific
                              log-likelihoods
        """

        super(GWRResults, self).__init__(
            X, y, kernel, fixed, influ, reside, predict_value, betas, tr_STS)
        self.coords = coords
        self.bw = bw
        self.CCT = CCT * self.sigma2
        self.bse = np.sqrt(self.CCT)
        self.tvalues = self.betas / self.bse


class GTWRResults(Results):

    def __init__(
            self, coords, t, X, y, bw, tau, kernel, fixed, influ, reside, predict_value, betas, CCT, tr_STS
    ):
        """
        tau:        : scalar
                      spatio-temporal scale
        bw_s        : scalar
                      spatial bandwidth
        bw_t        : scalar
                      temporal bandwidth
        See Also GWRResults
        """

        super(GTWRResults, self).__init__(X, y, kernel, fixed, influ, reside, predict_value, betas, tr_STS)
        self.coords = coords
        self.t = t
        self.bw = bw
        self.tau = tau
        self.bw_s = self.bw
        self.bw_t = np.sqrt(self.bw ** 2 / self.tau)
        self.CCT = CCT * self.sigma2
        self.bse = np.sqrt(self.CCT)
        self.tvalues = self.betas / self.bse


class MGWRResults(BaseModel):

    def __init__(self, coords, X, y, bws, kernel, fixed, bws_history, betas,
                 predict_value, ENP_j, CCT):
        """
        bws         : array-like
                      corresponding spatial bandwidth of all variables
        ENP_j       : array-like
                      effective number of paramters, which depends on
                      sigma2, for each covariate in the model

        See Also GWRResults
        """
        super(MGWRResults, self).__init__(X, y, kernel, fixed, constant=False)
        self.coords = coords
        self.bws = bws
        self.bws_history = bws_history
        self.predict_value = predict_value
        self.betas = betas
        self.ENP_j = ENP_j
        self.reside = self.y - self.predict_value
        self.tr_S = np.sum(self.ENP_j)
        self.ENP = self.tr_S
        self.TSS = np.sum((self.y - np.mean(self.y)) ** 2)
        self.RSS = np.sum(self.reside ** 2)
        self.sigma2 = self.RSS / (self.n - self.tr_S)
        self.CCT = CCT * self.sigma2
        self.bse = np.sqrt(self.CCT)
        self.t_values = self.betas / self.bse
        self.df_model = self.n - self.tr_S
        self.R2 = 1 - self.RSS / self.TSS
        self.adj_R2 = 1 - (1 - self.R2) * (self.n - 1) / (self.n - self.ENP - 1)
        self.llf = -np.log(self.RSS) * self.n / \
                   2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
        self.aic = -2.0 * self.llf + 2.0 * (self.tr_S + 1)
        self.aic_c = self.aic + 2.0 * self.tr_S * (self.tr_S + 1.0) / \
                     (self.n - self.tr_S - 1.0)
        self.bic = -2.0 * self.llf + (self.k + 1) * np.log(self.n)


class MGTWRResults(MGWRResults):

    def __init__(self, coords, t, X, y, bws, taus, kernel, fixed, bw_ts, bws_history, taus_history, betas,
                 predict_value, ENP_j, CCT):
        """
        taus        : array-like
                     corresponding spatio-temporal scale of all variables
        bws         : array-like
                     corresponding spatio bandwidth of all variables
        bw_ts       : array-like
                     corresponding temporal bandwidth of all variables
        See Also
        -------------
        MGWRResults
        GWRResults
        """
        super(MGTWRResults, self).__init__(
            coords, X, y, bws, kernel, fixed, bws_history, betas, predict_value, ENP_j, CCT)
        self.t = t
        self.taus = taus
        self.bw_ts = bw_ts
        self.taus_history = taus_history
